@@ -8,59 +8,17 @@ import re
 import os
 import sys
 from shlex import split as shsplit
+try:
+    from itertools import zip_longest
+except ImportError:
+    from itertools import izip_longest as zip_longest  # Python 2
 
-import vim
-import jedi
 
 is_py3 = sys.version_info[0] >= 3
 if is_py3:
     unicode = str
 
 
-def catch_and_print_exceptions(func):
-    def wrapper(*args, **kwargs):
-        try:
-            return func(*args, **kwargs)
-        except (Exception, vim.error):
-            print(traceback.format_exc())
-            return None
-    return wrapper
-
-
-class VimError(Exception):
-    def __init__(self, message, throwpoint, executing):
-        super(type(self), self).__init__(message)
-        self.throwpoint = throwpoint
-        self.executing = executing
-
-    def __str__(self):
-        return self.message + '; created by: ' + repr(self.executing)
-
-
-def _catch_exception(string, is_eval):
-    """
-    Interface between vim and python calls back to it.
-    Necessary, because the exact error message is not given by `vim.error`.
-    """
-    e = 'jedi#_vim_exceptions(%s, %s)'
-    result = vim.eval(e % (repr(PythonToVimStr(string, 'UTF-8')), is_eval))
-    if 'exception' in result:
-        raise VimError(result['exception'], result['throwpoint'], string)
-    return result['result']
-
-
-def vim_eval(string):
-    return _catch_exception(string, 1)
-
-
-def vim_command(string):
-    _catch_exception(string, 0)
-
-
-def echo_highlight(msg):
-    vim_command('echohl WarningMsg | echom "%s" | echohl None' % msg)
-
-
 class PythonToVimStr(unicode):
     """ Vim has a different string implementation of single quotes """
     __slots__ = []
@@ -83,6 +41,92 @@ class PythonToVimStr(unicode):
         return '"%s"' % s.replace('\\', '\\\\').replace('"', r'\"')
 
 
+class VimError(Exception):
+    def __init__(self, message, throwpoint, executing):
+        super(type(self), self).__init__(message)
+        self.message = message
+        self.throwpoint = throwpoint
+        self.executing = executing
+
+    def __str__(self):
+        return self.message + '; created by: ' + repr(self.executing)
+
+
+def _catch_exception(string, is_eval):
+    """
+    Interface between vim and python calls back to it.
+    Necessary, because the exact error message is not given by `vim.error`.
+    """
+    e = 'jedi#_vim_exceptions(%s, %s)'
+    result = vim.eval(e % (repr(PythonToVimStr(string, 'UTF-8')), is_eval))
+    if 'exception' in result:
+        raise VimError(result['exception'], result['throwpoint'], string)
+    return result['result']
+
+
+def vim_command(string):
+    _catch_exception(string, 0)
+
+
+def vim_eval(string):
+    return _catch_exception(string, 1)
+
+
+def no_jedi_warning():
+    vim.command('echohl WarningMsg'
+                '| echom "Please install Jedi if you want to use jedi-vim."'
+                '| echohl None')
+
+
+def echo_highlight(msg):
+    vim_command('echohl WarningMsg | echom "{}" | echohl None'.format(
+        msg.replace('"', '\\"')))
+
+
+import vim
+try:
+    import jedi
+except ImportError:
+    no_jedi_warning()
+    jedi = None
+else:
+    try:
+        version = jedi.__version__
+    except Exception as e:  # e.g. AttributeError
+        echo_highlight("Could not load jedi python module: {}".format(e))
+        jedi = None
+    else:
+        if isinstance(version, str):
+            # the normal use case, now.
+            from jedi import utils
+            version = utils.version_info()
+        if version < (0, 7):
+            echo_highlight('Please update your Jedi version, it is too old.')
+
+
+def catch_and_print_exceptions(func):
+    def wrapper(*args, **kwargs):
+        try:
+            return func(*args, **kwargs)
+        except (Exception, vim.error):
+            print(traceback.format_exc())
+            return None
+    return wrapper
+
+
+def _check_jedi_availability(show_error=False):
+    def func_receiver(func):
+        def wrapper(*args, **kwargs):
+            if jedi is None:
+                if show_error:
+                    no_jedi_warning()
+                return
+            else:
+                return func(*args, **kwargs)
+        return wrapper
+    return func_receiver
+
+
 @catch_and_print_exceptions
 def get_script(source=None, column=None):
     jedi.settings.additional_dynamic_modules = \
@@ -97,10 +141,14 @@ def get_script(source=None, column=None):
     return jedi.Script(source, row, column, buf_path, encoding)
 
 
+@_check_jedi_availability(show_error=False)
 @catch_and_print_exceptions
 def completions():
     row, column = vim.current.window.cursor
-    clear_call_signatures()
+    # Clear call signatures in the buffer so they aren't seen by the completer.
+    # Call signatures in the command line can stay.
+    if vim_eval("g:jedi#show_call_signatures") == '1':
+        clear_call_signatures()
     if vim.eval('a:findstart') == '1':
         count = 0
         for char in reversed(vim.current.line[:column]):
@@ -149,25 +197,36 @@ def completions():
         vim.command('return ' + strout)
 
 
+@_check_jedi_availability(show_error=True)
 @catch_and_print_exceptions
-def goto(is_definition=False, is_related_name=False, no_output=False):
-    definitions = []
+def goto(mode="goto", no_output=False):
+    """
+    :param str mode: "related_name", "definition", "assignment", "auto"
+    :return: list of definitions/assignments
+    :rtype: list
+    """
     script = get_script()
     try:
-        if is_related_name:
+        if mode == "goto":
+            definitions = [x for x in script.goto_definitions()
+                           if not x.in_builtin_module()]
+            if not definitions:
+                definitions = script.goto_assignments()
+        elif mode == "related_name":
             definitions = script.usages()
-        elif is_definition:
+        elif mode == "definition":
             definitions = script.goto_definitions()
-        else:
+        elif mode == "assignment":
             definitions = script.goto_assignments()
     except jedi.NotFoundError:
         echo_highlight("Cannot follow nothing. Put your cursor on a valid name.")
+        definitions = []
     else:
         if no_output:
             return definitions
         if not definitions:
             echo_highlight("Couldn't find any definitions for this.")
-        elif len(definitions) == 1 and not is_related_name:
+        elif len(definitions) == 1 and mode != "related_name":
             # just add some mark to add the current position to the jumplist.
             # this is ugly, because it overrides the mark for '`', so if anyone
             # has a better idea, let me know.
@@ -179,14 +238,13 @@ def goto(is_definition=False, is_related_name=False, no_output=False):
                     echo_highlight("Cannot get the definition of Python keywords.")
                 else:
                     echo_highlight("Builtin modules cannot be displayed (%s)."
-                                   % d.module_path)
+                                   % d.desc_with_module)
             else:
                 if d.module_path != vim.current.buffer.name:
                     result = new_buffer(d.module_path)
                     if not result:
-                        return
+                        return []
                 vim.current.window.cursor = d.line, d.column
-                vim_command('normal! zt')  # cursor at top of screen
         else:
             # multiple solutions
             lst = []
@@ -198,10 +256,11 @@ def goto(is_definition=False, is_related_name=False, no_output=False):
                                     lnum=d.line, col=d.column + 1,
                                     text=PythonToVimStr(d.description)))
             vim_eval('setqflist(%s)' % repr(lst))
-            vim_eval('jedi#add_goto_window()')
+            vim_eval('jedi#add_goto_window(' + str(len(lst)) + ')')
     return definitions
 
 
+@_check_jedi_availability(show_error=True)
 @catch_and_print_exceptions
 def show_documentation():
     script = get_script()
@@ -224,25 +283,35 @@ def show_documentation():
         text = ('\n' + '-' * 79 + '\n').join(docs)
         vim.command('let l:doc = %s' % repr(PythonToVimStr(text)))
         vim.command('let l:doc_lines = %s' % len(text.split('\n')))
+    return True
 
 
 @catch_and_print_exceptions
 def clear_call_signatures():
+    # Check if using command line call signatures
+    if vim_eval("g:jedi#show_call_signatures") == '2':
+        vim_command('echo ""')
+        return
     cursor = vim.current.window.cursor
     e = vim_eval('g:jedi#call_signature_escape')
-    regex = r'%sjedi=([0-9]+), ([^%s]*)%s.*%sjedi%s'.replace('%s', e)
+    # We need two turns here to search and replace certain lines:
+    # 1. Search for a line with a call signature and save the appended
+    #    characters
+    # 2. Actually replace the line and redo the status quo.
+    py_regex = r'%sjedi=([0-9]+), (.*?)%s.*?%sjedi%s'.replace('%s', e)
     for i, line in enumerate(vim.current.buffer):
-        match = re.search(r'%s' % regex, line)
+        match = re.search(py_regex, line)
         if match is not None:
-            vim_regex = r'\v' + regex.replace('=', r'\=') + '.{%s}' \
-                % int(match.group(1))
-            vim_command(r'try | %s,%ss/%s/\2/g | catch | endtry'
-                        % (i + 1, i + 1, vim_regex))
-            vim_eval('histdel("search", -1)')
-            vim_command('let @/ = histget("search", -1)')
+            # Some signs were added to minimize syntax changes due to call
+            # signatures. We have to remove them again. The number of them is
+            # specified in `match.group(1)`.
+            after = line[match.end() + int(match.group(1)):]
+            line = line[:match.start()] + match.group(2) + after
+            vim.current.buffer[i] = line
     vim.current.window.cursor = cursor
 
 
+@_check_jedi_availability(show_error=False)
 @catch_and_print_exceptions
 def show_call_signatures(signatures=()):
     if vim_eval("has('conceal') && g:jedi#show_call_signatures") == '0':
@@ -255,6 +324,9 @@ def show_call_signatures(signatures=()):
     if not signatures:
         return
 
+    if vim_eval("g:jedi#show_call_signatures") == '2':
+        return cmdline_call_signatures(signatures)
+
     for i, signature in enumerate(signatures):
         line, column = signature.bracket_start
         # signatures are listed above each other
@@ -270,7 +342,8 @@ def show_call_signatures(signatures=()):
 
         params = [p.description.replace('\n', '') for p in signature.params]
         try:
-            params[signature.index] = '*%s*' % params[signature.index]
+            # *_*PLACEHOLDER*_* makes something fat. See after/syntax file.
+            params[signature.index] = '*_*%s*_*' % params[signature.index]
         except (IndexError, TypeError):
             pass
 
@@ -310,58 +383,151 @@ def show_call_signatures(signatures=()):
         vim_eval('setline(%s, %s)' % (line_to_replace, repr(PythonToVimStr(repl))))
 
 
+@catch_and_print_exceptions
+def cmdline_call_signatures(signatures):
+    def get_params(s):
+        return [p.description.replace('\n', '') for p in s.params]
+
+    if len(signatures) > 1:
+        params = zip_longest(*map(get_params, signatures), fillvalue='_')
+        params = ['(' + ', '.join(p) + ')' for p in params]
+    else:
+        params = get_params(signatures[0])
+    text = ', '.join(params).replace('"', '\\"').replace(r'\n', r'\\n')
+
+    # Allow 12 characters for ruler/showcmd - setting noruler/noshowcmd
+    # here causes incorrect undo history
+    max_msg_len = int(vim_eval('&columns')) - 12
+    max_num_spaces = (max_msg_len - len(signatures[0].call_name)
+                      - len(text) - 2)  # 2 accounts for parentheses
+    if max_num_spaces < 0:
+        return  # No room for the message
+    _, column = signatures[0].bracket_start
+    num_spaces = min(int(vim_eval('g:jedi#first_col +'
+                     'wincol() - col(".")')) +
+                     column - len(signatures[0].call_name),
+                     max_num_spaces)
+    spaces = ' ' * num_spaces
+
+    try:
+        index = [s.index for s in signatures if isinstance(s.index, int)][0]
+        left = text.index(params[index])
+        right = left + len(params[index])
+        vim_command('                      echon "%s" | '
+                    'echohl Function     | echon "%s" | '
+                    'echohl None         | echon "("  | '
+                    'echohl jediFunction | echon "%s" | '
+                    'echohl jediFat      | echon "%s" | '
+                    'echohl jediFunction | echon "%s" | '
+                    'echohl None         | echon ")"'
+                    % (spaces, signatures[0].call_name, text[:left],
+                       text[left:right], text[right:]))
+    except (TypeError, IndexError):
+        vim_command('                      echon "%s" | '
+                    'echohl Function     | echon "%s" | '
+                    'echohl None         | echon "("  | '
+                    'echohl jediFunction | echon "%s" | '
+                    'echohl None         | echon ")"'
+                    % (spaces, signatures[0].call_name, text))
+
+
+@_check_jedi_availability(show_error=True)
 @catch_and_print_exceptions
 def rename():
     if not int(vim.eval('a:0')):
-        _rename_cursor = vim.current.window.cursor
-
-        vim_command('normal A ')  # otherwise startinsert doesn't work well
-        vim.current.window.cursor = _rename_cursor
-
         vim_command('augroup jedi_rename')
         vim_command('autocmd InsertLeave <buffer> call jedi#rename(1)')
         vim_command('augroup END')
 
+        vim_command("let s:jedi_replace_orig = expand('<cword>')")
         vim_command('normal! diw')
-        vim_command(':startinsert')
+        vim_command("let s:jedi_changedtick = b:changedtick")
+        vim_command('startinsert')
+
     else:
-        window_path = vim.current.buffer.name
-        # reset autocommand
+        # Remove autocommand.
         vim_command('autocmd! jedi_rename InsertLeave')
 
-        replace = vim_eval("expand('<cword>')")
-        vim_command('normal! u')  # undo new word
+        # Get replacement, if there is something on the cursor.
+        # This won't be the case when the user ends insert mode right away,
+        # and `<cword>` would pick up the nearest word instead.
+        if vim_eval('getline(".")[getpos(".")[2]-1]') != ' ':
+            replace = vim_eval("expand('<cword>')")
+        else:
+            replace = None
+
         cursor = vim.current.window.cursor
-        vim_command('normal! u')  # undo the space at the end
+
+        # Undo new word, but only if something was changed, which is not the
+        # case when ending insert mode right away.
+        if vim_eval('b:changedtick != s:jedi_changedtick') == '1':
+            vim_command('normal! u')  # Undo new word.
+        vim_command('normal! u')  # Undo diw.
+
         vim.current.window.cursor = cursor
 
-        if replace is None:
-            echo_highlight('No rename possible, if no name is given.')
-        else:
-            temp_rename = goto(is_related_name=True, no_output=True)
-            # sort the whole thing reverse (positions at the end of the line
-            # must be first, because they move the stuff before the position).
-            temp_rename = sorted(temp_rename, reverse=True,
-                                 key=lambda x: (x.module_path, x.start_pos))
-            for r in temp_rename:
-                if r.in_builtin_module():
-                    continue
+        if replace:
+            return do_rename(replace)
 
-                if vim.current.buffer.name != r.module_path:
-                    result = new_buffer(r.module_path)
-                    if not result:
-                        return
 
-                vim.current.window.cursor = r.start_pos
-                vim_command('normal! cw%s' % replace)
+def rename_visual():
+    replace = vim.eval('input("Rename to: ")')
+    orig = vim.eval('getline(".")[(getpos("\'<")[2]-1):getpos("\'>")[2]]')
+    do_rename(replace, orig)
 
-            result = new_buffer(window_path)
+
+def do_rename(replace, orig=None):
+    if not len(replace):
+        echo_highlight('No rename possible without name.')
+        return
+
+    if orig is None:
+        orig = vim_eval('s:jedi_replace_orig')
+
+    # Save original window / tab.
+    saved_tab = int(vim_eval('tabpagenr()'))
+    saved_win = int(vim_eval('winnr()'))
+
+    temp_rename = goto(mode="related_name", no_output=True)
+    # Sort the whole thing reverse (positions at the end of the line
+    # must be first, because they move the stuff before the position).
+    temp_rename = sorted(temp_rename, reverse=True,
+                         key=lambda x: (x.module_path, x.start_pos))
+    buffers = set()
+    for r in temp_rename:
+        if r.in_builtin_module():
+            continue
+
+        if os.path.abspath(vim.current.buffer.name) != r.module_path:
+            result = new_buffer(r.module_path)
             if not result:
-                return
-            vim.current.window.cursor = cursor
-            echo_highlight('Jedi did %s renames!' % len(temp_rename))
+                echo_highlight("Jedi-vim: failed to create buffer window for {}!".format(r.module_path))
+                continue
+
+        buffers.add(vim.current.buffer.name)
+
+        # Save view.
+        saved_view = vim_eval('string(winsaveview())')
+
+        # Replace original word.
+        vim.current.window.cursor = r.start_pos
+        vim_command('normal! c{:d}l{}'.format(len(orig), replace))
+
+        # Restore view.
+        vim_command('call winrestview(%s)' % saved_view)
+
+    # Restore previous tab and window.
+    vim_command('tabnext {:d}'.format(saved_tab))
+    vim_command('{:d}wincmd w'.format(saved_win))
+
+    if len(buffers) > 1:
+        echo_highlight('Jedi did {:d} renames in {:d} buffers!'.format(
+            len(temp_rename), len(buffers)))
+    else:
+        echo_highlight('Jedi did {:d} renames!'.format(len(temp_rename)))
 
 
+@_check_jedi_availability(show_error=True)
 @catch_and_print_exceptions
 def py_import():
     # args are the same as for the :edit command
@@ -407,8 +573,11 @@ def new_buffer(path, options=''):
             'top': 'topleft split',
             'left': 'topleft vsplit',
             'right': 'botright vsplit',
-            'bottom': 'botright split'
+            'bottom': 'botright split',
+            'winwidth': 'vs'
         }
+        if user_split_option == 'winwidth' and vim.current.window.width <= 2 * int(vim_eval("&textwidth ? &textwidth : 80")):
+            split_options['winwidth'] = 'sp'
         if user_split_option not in split_options:
             print('g:jedi#use_splits_not_buffers value is not correct, valid options are: %s' % ','.join(split_options.keys()))
         else:
@@ -454,6 +623,8 @@ def _tabnew(path, options=''):
                 if buf_path == path:
                     # tab exists, just switch to that tab
                     vim_command('tabfirst | tabnext %i' % (tab_nr + 1))
+                    # Goto the buffer's window.
+                    vim_command('exec bufwinnr(%i) . " wincmd w"' % (buf_nr + 1))
                     break
         else:
             continue
@@ -469,12 +640,3 @@ def escape_file_path(path):
 
 def print_to_stdout(level, str_out):
     print(str_out)
-
-
-version = jedi.__version__
-if isinstance(version, str):
-    # the normal use case, now.
-    from jedi import utils
-    version = utils.version_info()
-if version < (0, 7):
-    echo_highlight('Please update your Jedi version, it is to old.')
